#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys\types.h> 
#include <sys\stat.h>

#include "tap.h"

unsigned char* _tapFileBuffer = NULL;
unsigned int _tapFileSize = 0;
unsigned char _tapLoadedFile = 0;
char* _tapLoadedFileName = NULL;

#define _HEADER_FILENAME_LENGTH 10

#define _MAX_TAP_BLOCKS 1000
struct tapBlock** _tapBlocks = NULL;
unsigned short _tapBlocksCount = 0;

const char* _tap_translate_block_type(unsigned char flagByte) {
    switch (flagByte) {
    case 0x00:
        return "[hdr]";
    case 0xFF:
        return "[dat]";
    default:
        return "[???]";
    }
}

void tap_start() {
}

void tap_destroy() {
    
    if (_tapFileBuffer != NULL) {
        free(_tapFileBuffer);
    }

    // free TAP blocks array
    if (_tapBlocks) {
        for (int i = 0; i < _tapBlocksCount; i++) {
            struct tapBlock* block = _tapBlocks[i];

            // Don't need MSVC's antics here
#pragma warning(push)
#pragma warning(disable: 6001)
            if (block->type == Header) {
                free(block->headerInfo->filename);
                free(block->headerInfo);
            }
            free(block);
#pragma warning(pop)
        }
        free(_tapBlocks);
    }
}

unsigned short tap_get_block_count() {
    return _tapBlocksCount;
}

// reads a header-type block
// tapBuffer is a ptr to a ptr to block's second byte (aka immediately after flag)
struct tapBlock* _tap_readblock__header(unsigned char** tapBuffer, unsigned short blockSize, unsigned char flagByte) {
    unsigned char* ptrToHeaderByte0 = *tapBuffer;

    // header byte 0: block type: 0=Program, 1=Number array, 2=Character array, 3=Code
    unsigned char blockType = **tapBuffer;
    (*tapBuffer)++;

    // header bytes 1-10: file name, padded with 0x20
    char* filename = (char*)malloc((_HEADER_FILENAME_LENGTH + 1) * sizeof(char) + 1);
    if (filename) {
        memcpy(filename, (char*)(*tapBuffer), _HEADER_FILENAME_LENGTH);
        filename[_HEADER_FILENAME_LENGTH] = 0; // terminate
    }
    (*tapBuffer) += _HEADER_FILENAME_LENGTH;

    // header bytes 11-12: length of data block
    unsigned short dataBlockLength = (**(unsigned short**)tapBuffer);
    (*tapBuffer) += 2;

    // header bytes 13-14: parameter1
    unsigned short parameter1 = (**(unsigned short**)tapBuffer);
    (*tapBuffer) += 2;

    // header bytes 15-16: parameter2
    unsigned short parameter2 = (**(unsigned short**)tapBuffer);
    (*tapBuffer) += 2;

    // byte 0 after header: checksum
    unsigned char checksum = **tapBuffer;
    (*tapBuffer)++;

    // *tapBuffer now points to first byte immediately after block

    // build result
    struct tapHeaderBlock* header = (struct tapHeaderBlock*)malloc(1 * sizeof(struct tapHeaderBlock));
    struct tapBlock* wrapper = (struct tapBlock*)malloc(1 * sizeof(struct tapBlock));
    if (header && wrapper) {
        header->type = (enum tapHeaderType)blockType;
        header->filename = filename;
        header->dataBlockLength = dataBlockLength;
        header->parameter1 = parameter1;
        header->parameter2 = parameter2;

        wrapper->type = Header;
        wrapper->flagByte = flagByte;
        wrapper->headerInfo = header;
        wrapper->tapBlockSize = blockSize + 2;      // plus 2 initial TAP-specific size bytes (unknown to ZX Spectrum)
        wrapper->blockDataSize = blockSize;
        wrapper->blockRawDataSize = blockSize - 2;  // minus 2 ZX Spectrum metadata bytes (block type, checksum)
        wrapper->tapBlockBytes = ptrToHeaderByte0 - 3;  // rewind 3 bytes to point to first size byte
        wrapper->tapBlockOffset = wrapper->tapBlockBytes - _tapFileBuffer;
        wrapper->tapBlockEndOffset = wrapper->tapBlockOffset + wrapper->tapBlockSize - 1;
    }
    return wrapper;
}

// reads a data-type block
// tapBuffer is a ptr to a ptr to block's second byte (aka immediately after flag)
struct tapBlock* _tap_readblock__data(unsigned char** tapBuffer, unsigned short blockSize, unsigned char flagByte) {
    unsigned char* ptrToDataByte0 = *tapBuffer;

    (*tapBuffer) += blockSize - 1;  // - 1 because we're already on the second byte of the block
    // *tapBuffer now points to first byte immediately after block

    // build result
    struct tapBlock* result = (struct tapBlock*)malloc(1 * sizeof(struct tapBlock));
    if (result) {
        result->blockRawDataSize = blockSize - 2; // minus 2 ZX Spectrum metadata bytes (block type, checksum)
        result->blockDataSize = blockSize;
        result->tapBlockSize = blockSize + 2;     // plus 2 initial size bytes
        result->tapBlockBytes = ptrToDataByte0 - 3;  // rewind 3 bytes to point to first size byte
        result->headerInfo = NULL;
        result->type = Data;
        result->flagByte = flagByte;
        result->tapBlockOffset = result->tapBlockBytes - _tapFileBuffer;
        result->tapBlockEndOffset = result->tapBlockOffset + result->tapBlockSize - 1;
    }
    return result;
}

// assumes *tapBuffer points to the start of a block
// moves *tapBuffer to immediately after the block it reads
struct tapBlock* _tap_readblock(unsigned char** tapBuffer) {
    // bytes 0-1: size of header, low endian
    unsigned short blockSize = (**(unsigned short**)tapBuffer);
    (*tapBuffer) += 2;

    // byte 2: flag byte, 0=header, 0xFF=body of file
    unsigned char flagByte = **tapBuffer;
    (*tapBuffer)++;

    struct tapBlock* block = NULL;
    switch (flagByte)
    {
    case 0:
        block = _tap_readblock__header(tapBuffer, blockSize, flagByte);
        break;
    default:
        block = _tap_readblock__data(tapBuffer, blockSize, flagByte);
        break;
    }
    block->flagByteHumanReadable = _tap_translate_block_type(block->flagByte);

    return block;
}

void _tap_enumerate_blocks() {
    _tapBlocks = (struct tapBlock**)malloc(_MAX_TAP_BLOCKS * sizeof(struct tapBlock*));
    if (!_tapBlocks) {
        printf("Error: unable to allocate memory for TAP block array\n");
        return;
    }

    printf("Slot\tType\tFlag\tName\t\tStart\t\tLength\tEnd\n");
    printf("----\t----\t----\t----\t\t-----\t\t------\t---\n");
    unsigned char* cursor = _tapFileBuffer;
    while (_tapBlocksCount < _MAX_TAP_BLOCKS) {
        if ((long long)cursor - (long long)_tapFileBuffer >= (long long)_tapFileSize) {
            break;
        }
        struct tapBlock* block = _tap_readblock(&cursor);
        _tapBlocks[_tapBlocksCount] = block;
        block->blockIndex = _tapBlocksCount;

        printf("%u\t%s\t0x%02X\t%s\t0x%08X\t0x%04X\t0x%08X\n",
            _tapBlocksCount,
            block->flagByteHumanReadable,
            block->flagByte,
            block->type == Header ? block->headerInfo->filename : "\t",
            block->tapBlockOffset,
            block->tapBlockSize,
            block->tapBlockEndOffset);

        _tapBlocksCount++;
    }

    if (_tapBlocksCount == _MAX_TAP_BLOCKS) {
        printf("Warning: max block count exceeded in TAP; loading only first %u blocks\n", _MAX_TAP_BLOCKS);
    }
}

unsigned char tap_load(char* filename) {
    struct stat info;
    if (stat(filename, &info) != 0) {
        printf("Error: TAP file not found\n");
        return 0;
    }

    _tapFileSize = info.st_size;
    if (_tapFileSize < 3) {
        printf("Error: TAP file must be at least 3 bytes long\n");
        return 0;
    }

    _tapFileBuffer = (unsigned char*)malloc(_tapFileSize);
    if (_tapFileBuffer == NULL) {
        printf("Error: unable to allocate memory for TAP file\n");
        return 0;
    }
    FILE* fp;
    errno_t result = fopen_s(&fp, filename, "rb");
    if (result) {
        printf("Error: could not open TAP file\n");
        return 0;
    }
    // try to read a single block of info.st_size bytes
    size_t blocks_read = fread(_tapFileBuffer, _tapFileSize, 1, fp);
    if (blocks_read != 1) {
        printf("Error: could not read from TAP file\n");
        return 0;
    }
    fclose(fp);

    printf("Loaded TAP file of %u bytes (0x%08X)\n", _tapFileSize, _tapFileSize);

    _tap_enumerate_blocks();
    _tapLoadedFile = 1;
    _tapLoadedFileName = filename;
    return 1;
}

unsigned char tap_is_file_loaded() {
    return _tapLoadedFile;
}

char* tap_get_filename() {
    return _tapLoadedFileName;
}

struct tapBlock* tap_get_block(int index) {
    return _tapBlocks[index];
}
